package ru.landgrafhomyak.serdha.api.v0.ddl.module

import ru.landgrafhomyak.serdha.api.v0.ddl.table.Table
import ru.landgrafhomyak.serdha.api.v0.dml.CopyQuery
import ru.landgrafhomyak.serdha.api.v0.dml.CreateCopyQueryScope
import ru.landgrafhomyak.serdha.api.v0.dml.CreateInsertQueryScope
import ru.landgrafhomyak.serdha.api.v0.dml.DeleteQuery
import ru.landgrafhomyak.serdha.api.v0.dml.InsertQuery
import ru.landgrafhomyak.serdha.api.v0.dml.SelectQuery
import ru.landgrafhomyak.serdha.api.v0.dml.UpdateQuery
import ru.landgrafhomyak.serdha.api.v0.runtime.SynchronizedDatabase
import ru.landgrafhomyak.serdha.api.v0.runtime.Transaction

/**
 * Scope methods for creating new [module template][ModuleTemplate].
 *
 * @see ModuleTemplate.CreateModuleTemplatesScope.createTemplate
 * @see ModuleTemplate.ModuleCreator.createSchema
 */
public interface CreateModuleScope {
	/**
	 * Functional interface for creating table.
	 *
	 * Can be implemented like a companion object that just calls [TableUserExtension]'s constructor with the same or similar signature.
	 *
	 * @param TableUserExtension User's type for containing table-related descriptors.
	 *
	 * @see CreateModuleScope.createTable
	 * @see CreateModuleScope.createSessionScopeTemporaryTable
	 * @see CreateModuleScope.createTransactionScopeTemporaryTable
	 * @see UpgradeModuleScope.TableUpgrade
	 */
	public interface TableCreator<TableUserExtension : Any> {
		/**
		 * Scope method that uses [creator] to initialize table schema.
		 *
		 * @param creator Object with descriptor providers and factories.
		 * @return User's object with descriptors for future access.
		 *
		 * @see CreateModuleScope.createTable
		 * @see CreateModuleScope.createSessionScopeTemporaryTable
		 * @see CreateModuleScope.createTransactionScopeTemporaryTable
		 * @see UpgradeModuleScope.TableUpgrade
		 */
		public fun createTable(creator: ru.landgrafhomyak.serdha.api.v0.ddl.table.CreateTableScope<TableUserExtension>): TableUserExtension
	}

	/**
	 * Creates table in [specified namespace][namespace] with [specified name][name] and stores it in the database.
	 * If table was created by previous synchronizations, asserts that table at the specified path in the database is same as table provided by [initializer].
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] this table must be [kept][UpgradeModuleScope.upgradeTable] (or [renamed][CreateModuleScope.renameTable]) or [deleted][CreateModuleScope.deleteTable].
	 *
	 * @param TableUserExtension User's type for containing table-related descriptors.
	 * @param namespace Way to group tables if there is a lot.
	 * @param name Name of table for debugging.
	 * @param initializer Table-related descriptors initializer.
	 * @return Table descriptor.
	 */
	public fun <TableUserExtension : Any> createTable(namespace: Namespace, name: String, initializer: TableCreator<TableUserExtension>): Table<TableUserExtension, Nothing>

	/**
	 * Creates a temporary table that exits only until connection to the database not [closed][SynchronizedDatabase._close].
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TableUserExtension User's type for containing table-related descriptors.
	 * @param namespace Way to group tables if there is a lot.
	 * @param name Name of table for debugging.
	 * @param initializer Table-related descriptors initializer.
	 * @return Table descriptor.
	 */
	public fun <TableUserExtension : Any> createSessionScopeTemporaryTable(namespace: Namespace, name: String, initializer: TableCreator<TableUserExtension>): Table<TableUserExtension, Nothing>

	/**
	 * Creates a temporary table that exits only inside [transaction][Transaction] and auto-deleted (or cleared, depends on driver implementation) when it [finished][Transaction._assertTransactionFinishedAndReleaseResources].
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TableUserExtension User's type for containing table-related descriptors.
	 * @param namespace Way to group tables if there is a lot.
	 * @param name Name of table for debugging.
	 * @param initializer Table-related descriptors initializer.
	 * @return Table descriptor.
	 */
	public fun <TableUserExtension : Any> createTransactionScopeTemporaryTable(namespace: Namespace, name: String, initializer: TableCreator<TableUserExtension>): Table<TableUserExtension, Nothing>

	/**
	 * Functional interface for creating 'SELECT' query.
	 *
	 * Can be implemented like a companion object that just calls [QueryUserWrapper]'s constructor with the same or similar signature.
	 *
	 * @param QueryUserWrapper User's type with descriptors related to query.
	 *
	 * @see CreateModuleScope.createSelectQuery
	 */
	public interface SelectQueryCreator<QueryUserWrapper : Any> {
		/**
		 * Scope method that uses [creator] to create a query.
		 *
		 * @param creator Object with descriptor providers and factories.
		 * @return User's object with query-related descriptors for future access.
		 *
		 * @see CreateModuleScope.createSelectQuery
		 */
		public fun createSelect(creator: ru.landgrafhomyak.serdha.api.v0.dml.CreateSelectQueryScope<QueryUserWrapper>): QueryUserWrapper
	}

	/**
	 * Creates 'SELECT' query.
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param QueryUserWrapper User's type for containing query-related descriptors.
	 * @param initializer Query-related descriptors initializer.
	 * @return Query descriptor.
	 *
	 * @see CreateModuleScope.createTemporarySelectQuery
	 */
	public fun <QueryUserWrapper : Any> createSelectQuery(initializer: SelectQueryCreator<QueryUserWrapper>): SelectQuery<QueryUserWrapper>

	/**
	 * Creates 'SELECT' query accessible only while the module is [creating][ModuleTemplate.ModuleCreator] or [updating][ModuleTemplate.ModuleUpgrade].
	 * Using this function instead of [CreateModuleScope.createSelectQuery] may help the driver to optimize caches and runtime indexes.
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param QueryUserWrapper User's type for containing query-related descriptors.
	 * @param initializer Query-related descriptors initializer.
	 * @return Query descriptor.
	 *
	 * @see CreateModuleScope.createSelectQuery
	 */
	public fun <QueryUserWrapper : Any> createTemporarySelectQuery(initializer: SelectQueryCreator<QueryUserWrapper>): SelectQuery<QueryUserWrapper>

	/**
	 * Functional interface for creating 'INSERT ... VALUES' query.
	 *
	 * Can be implemented like a companion object that just calls [QueryUserWrapper]'s constructor with the same or similar signature.
	 *
	 * @param TableUserExtension User's type with descriptors related to table to which values will be inserted.
	 * @param QueryUserWrapper User's type with descriptors related to query.
	 *
	 * @see CreateModuleScope.createInsertQuery
	 */
	public interface InsertQueryCreator<TableUserExtension : Any, QueryUserWrapper : Any> {
		/**
		 * Scope method that uses [creator] to create a query.
		 *
		 * @param table Descriptor of table to which values will be inserted.
		 * @param creator Object with descriptor providers and factories.
		 * @return User's object with query-related descriptors for future access.
		 *
		 * @see CreateModuleScope.createInsertQuery
		 */
		public fun createInsert(table: Table<TableUserExtension, *>, creator: CreateInsertQueryScope<TableUserExtension, QueryUserWrapper>): QueryUserWrapper
	}

	/**
	 * Creates 'INSERT ... VALUES' query.
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TableUserExtension User's type with descriptors related to table to which values will be inserted.
	 * @param QueryUserWrapper User's type for containing query-related descriptors.
	 * @param initializer Query-related descriptors initializer.
	 * @return Query descriptor.
	 *
	 * @see CreateModuleScope.createTemporaryInsertQuery
	 */
	public fun <TableUserExtension : Any, QueryUserWrapper : Any> createInsertQuery(table: Table<TableUserExtension, *>, initializer: InsertQueryCreator<TableUserExtension, QueryUserWrapper>): InsertQuery<QueryUserWrapper>

	/**
	 * Creates 'INSERT ... VALUES' query accessible only while the module is [creating][ModuleTemplate.ModuleCreator] or [updating][ModuleTemplate.ModuleUpgrade].
	 * Using this function instead of [CreateModuleScope.createInsertQuery] may help the driver to optimize caches and runtime indexes.
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TableUserExtension User's type with descriptors related to table to which values will be inserted.
	 * @param QueryUserWrapper User's type for containing query-related descriptors.
	 * @param initializer Query-related descriptors initializer.
	 * @return Query descriptor.
	 *
	 * @see CreateModuleScope.createInsertQuery
	 */
	public fun <TableUserExtension : Any, QueryUserWrapper : Any> createTemporaryInsertQuery(table: Table<TableUserExtension, *>, initializer: InsertQueryCreator<TableUserExtension, QueryUserWrapper>): InsertQuery<QueryUserWrapper>

	/**
	 * Functional interface for creating 'INSERT ... FROM' query.
	 *
	 * Can be implemented like a companion object that just calls [QueryUserWrapper]'s constructor with the same or similar signature.
	 *
	 * @param TableUserExtension User's type with descriptors related to table to which values will be inserted.
	 * @param QueryUserWrapper User's type with descriptors related to query.
	 *
	 * @see CreateModuleScope.createCopyQuery
	 */
	public interface CopyQueryCreator<TableUserExtension : Any, QueryUserWrapper : Any> {
		/**
		 * Scope method that uses [creator] to create a query.
		 *
		 * @param table Descriptor of table to which values will be inserted.
		 * @param creator Object with descriptor providers and factories.
		 * @return User's object with query-related descriptors for future access.
		 *
		 * @see CreateModuleScope.createCopyQuery
		 */
		public fun createCopy(table: Table<TableUserExtension, *>, creator: CreateCopyQueryScope<TableUserExtension, QueryUserWrapper>): QueryUserWrapper
	}

	/**
	 * Creates 'INSERT ... FROM' query.
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TableUserExtension User's type with descriptors related to table to which values will be inserted.
	 * @param initializer Query-related descriptors initializer.
	 * @return Query descriptor.
	 *
	 * @see CreateModuleScope.createTemporaryCopyQuery
	 */
	public fun <TableUserExtension : Any, QueryUserWrapper : Any> createCopyQuery(table: Table<TableUserExtension, *>, initializer: CopyQueryCreator<TableUserExtension, QueryUserWrapper>): CopyQuery<QueryUserWrapper>

	/**
	 * Creates 'INSERT ... FROM' query accessible only while the module is [creating][ModuleTemplate.ModuleCreator] or [updating][ModuleTemplate.ModuleUpgrade].
	 * Using this function instead of [CreateModuleScope.createCopyQuery] may help the driver to optimize caches and runtime indexes.
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TableUserExtension User's type with descriptors related to table to which values will be inserted.
	 * @param initializer Query-related descriptors initializer.
	 * @return Query descriptor.
	 *
	 * @see CreateModuleScope.createCopyQuery
	 */
	public fun <TableUserExtension : Any, QueryUserWrapper : Any> createTemporaryCopyQuery(table: Table<TableUserExtension, *>, initializer: CopyQueryCreator<TableUserExtension, QueryUserWrapper>): CopyQuery<QueryUserWrapper>

	/**
	 * Functional interface for creating 'UPDATE' query.
	 *
	 * Can be implemented like a companion object that just calls [QueryUserWrapper]'s constructor with the same or similar signature.
	 *
	 * @param TableUserExtension User's type with descriptors related to table, whose rows will be updated.
	 * @param QueryUserWrapper User's type with descriptors related to query.
	 *
	 * @see CreateModuleScope.createUpdateQuery
	 */
	public interface UpdateQueryCreator<TableUserExtension : Any, QueryUserWrapper : Any> {
		/**
		 * Scope method that uses [creator] to create a query.
		 *
		 * @param table Descriptor of table which rows will be updated.
		 * @param creator Object with descriptor providers and factories.
		 * @return User's object with query-related descriptors for future access.
		 *
		 * @see CreateModuleScope.createUpdateQuery
		 */
		public fun createUpdate(table: Table<TableUserExtension, *>, creator: ru.landgrafhomyak.serdha.api.v0.dml.CreateUpdateQueryScope<TableUserExtension, QueryUserWrapper>): QueryUserWrapper
	}

	/**
	 * Creates 'UPDATE' query.
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TableUserExtension User's type with descriptors related to table to which values will be inserted.
	 * @param initializer Query-related descriptors initializer.
	 * @return Query descriptor.
	 *
	 * @see CreateModuleScope.createTemporaryUpdateQuery
	 */
	public fun <TableUserExtension : Any, QueryUserWrapper : Any> createUpdateQuery(table: Table<TableUserExtension, *>, initializer: UpdateQueryCreator<TableUserExtension, QueryUserWrapper>): UpdateQuery<QueryUserWrapper>

	/**
	 * Creates 'UPDATE' query accessible only while the module is [creating][ModuleTemplate.ModuleCreator] or [updating][ModuleTemplate.ModuleUpgrade].
	 * Using this function instead of [CreateModuleScope.createUpdateQuery] may help the driver to optimize caches and runtime indexes.
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TableUserExtension User's type with descriptors related to table to which values will be inserted.
	 * @param initializer Query-related descriptors initializer.
	 * @return Query descriptor.
	 *
	 * @see CreateModuleScope.createUpdateQuery
	 */
	public fun <TableUserExtension : Any, QueryUserWrapper : Any> createTemporaryUpdateQuery(table: Table<TableUserExtension, *>, initializer: UpdateQueryCreator<TableUserExtension, QueryUserWrapper>): UpdateQuery<QueryUserWrapper>

	/**
	 * Functional interface for creating 'DELETE' query.
	 *
	 * Can be implemented like a companion object that just calls [QueryUserWrapper]'s constructor with the same or similar signature.
	 *
	 * @param TableUserExtension User's type with descriptors related to table from which data will be deleted.
	 * @param QueryUserWrapper User's type with descriptors related to query.
	 *
	 * @see CreateModuleScope.createDeleteQuery
	 */
	public interface DeleteQueryCreator<TableUserExtension : Any, QueryUserWrapper : Any> {
		/**
		 * Scope method that uses [creator] to create a query.
		 *
		 * @param table Descriptor of table from which data will be deleted.
		 * @param creator Object with descriptor providers and factories.
		 * @return User's object with query-related descriptors for future access.
		 *
		 * @see CreateModuleScope.createDeleteQuery
		 */
		public fun createDelete(table: Table<TableUserExtension, *>, creator: ru.landgrafhomyak.serdha.api.v0.dml.CreateDeleteQueryScope<TableUserExtension, QueryUserWrapper>): QueryUserWrapper
	}

	/**
	 * Creates 'DELETE' query.
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TableUserExtension User's type with descriptors related to table from which data will be deleted.
	 * @param initializer Query-related descriptors initializer.
	 * @return Query descriptor.
	 *
	 * @see CreateModuleScope.createTemporaryDeleteQuery
	 */
	public fun <TableUserExtension : Any, QueryUserWrapper : Any> createDeleteQuery(table: Table<TableUserExtension, *>, initializer: DeleteQueryCreator<TableUserExtension, QueryUserWrapper>): DeleteQuery<QueryUserWrapper>

	/**
	 * Creates 'DELETE' query accessible only while the module is [creating][ModuleTemplate.ModuleCreator] or [updating][ModuleTemplate.ModuleUpgrade].
	 * Using this function instead of [CreateModuleScope.createDeleteQuery] may help the driver to optimize caches and runtime indexes.
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TableUserExtension User's type with descriptors related to table from which data will be deleted.
	 * @param initializer Query-related descriptors initializer.
	 * @return Query descriptor.
	 *
	 * @see CreateModuleScope.createDeleteQuery
	 */
	public fun <TableUserExtension : Any, QueryUserWrapper : Any> createTemporaryDeleteQuery(table: Table<TableUserExtension, *>, initializer: DeleteQueryCreator<TableUserExtension, QueryUserWrapper>): DeleteQuery<QueryUserWrapper>

	/**
	 * Replaces [specified namespace][rootNs] with schema provided by [template].
	 * [This namespace][rootNs] will be root namespace for module and shouldn't contain any other definitions.
	 *
	 * @param rootNs Namespace where the new module will be located.
	 * @param template Schema template.
	 * @return Module descriptor.
	 */
	public fun <ModuleUserExtension : Any> substituteModule(rootNs: Namespace, template: ModuleTemplate<ModuleUserExtension>): Module<ModuleUserExtension>
}