package ru.landgrafhomyak.serdha.api.v0.ddl.types

import kotlin.jvm.JvmField
import kotlin.jvm.JvmName
import kotlinx.datetime.Instant
import kotlinx.datetime.LocalDate
import kotlinx.datetime.LocalDateTime
import kotlinx.datetime.LocalTime

/**
 * Descriptor of a database type.
 *
 * @param RuntimeType Type in programming language to which database type is converted.
 */
@Suppress("ClassName", "RemoveRedundantQualifierName")
public interface DatabaseType<@Suppress("unused") RuntimeType> {
	/**
	 * Name of database type for debugging and errors.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("name")
	public val name: String

	// todo
	public abstract class _VirtualType<R : Any>(@Suppress("MemberVisibilityCanBePrivate") @JvmField public val wraps: DatabaseType<R>)

	/**
	 * Descriptor of a nullable database type for static type-checking.
	 * @see DatabaseTypesProvider.nullableOf
	 */
	public interface Nullable<RT : Any, @Suppress("unused") DT : DatabaseType<RT>> : DatabaseType<RT?>

	/**
	 * Descriptor of an internal row id type for static type-checking.
	 * @param TargetTableUserWrapper  User's extension type for target table. For static type checking.
	 * @see DatabaseTypesProvider.ROW_ID
	 */
	public interface ROW_ID<TargetTableUserWrapper : Any> : DatabaseType<RowId<TargetTableUserWrapper>>

	/**
	 * Descriptor of a boolean type for static type-checking.
	 * @see DatabaseTypesProvider.BOOLEAN
	 */
	public interface BOOLEAN : DatabaseType<Boolean>

	/**
	 * Descriptor of an integer type for static type-checking.
	 * @see DatabaseTypesProvider.INT_S8
	 */
	public interface INT_S8 : DatabaseType<Byte>

	/**
	 * Descriptor of an integer type for static type-checking.
	 * @see DatabaseTypesProvider.INT_U8
	 */
	public interface INT_U8 : DatabaseType<UByte>

	/**
	 * Descriptor of an integer type for static type-checking.
	 * @see DatabaseTypesProvider.INT_S16
	 */
	public interface INT_S16 : DatabaseType<Short>

	/**
	 * Descriptor of an integer type for static type-checking.
	 * @see DatabaseTypesProvider.INT_U16
	 */
	public interface INT_U16 : DatabaseType<UShort>

	/**
	 * Descriptor of an integer type for static type-checking.
	 * @see DatabaseTypesProvider.INT_S32
	 */
	public interface INT_S32 : DatabaseType<Int>

	/**
	 * Descriptor of an integer type for static type-checking.
	 * @see DatabaseTypesProvider.INT_U32
	 */
	public interface INT_U32 : DatabaseType<UInt>

	/**
	 * Descriptor of an integer type for static type-checking.
	 * @see DatabaseTypesProvider.INT_S64
	 */
	public interface INT_S64 : DatabaseType<Long>

	/**
	 * Descriptor of an integer type for static type-checking.
	 * @see DatabaseTypesProvider.INT_U64
	 */
	public interface INT_U64 : DatabaseType<ULong>

	/**
	 * Descriptor of a date type for static type-checking.
	 * @see DatabaseTypesProvider.LOCAL_DATE
	 */
	public interface LOCAL_DATE : DatabaseType<LocalDate>

	/**
	 * Descriptor of a time type for static type-checking.
	 * @see DatabaseTypesProvider.LOCAL_TIME
	 */
	public interface LOCAL_TIME : DatabaseType<LocalTime>

	/**
	 * Descriptor of a date-time type for static type-checking.
	 * @see DatabaseTypesProvider.LOCAL_DATETIME
	 */
	public interface LOCAL_DATETIME : DatabaseType<LocalDateTime>

	/**
	 * Descriptor of a date-time type for static type-checking.
	 * @see DatabaseTypesProvider.INSTANT
	 */
	public interface INSTANT : DatabaseType<Instant>

	/**
	 * Descriptor of a floating-point number type for static type-checking.
	 * @see DatabaseTypesProvider.FLOAT_32
	 */
	public interface FLOAT_32 : DatabaseType<Float>

	/**
	 * Descriptor of a floating-point number type for static type-checking.
	 * @see DatabaseTypesProvider.FLOAT_64
	 */
	public interface FLOAT_64 : DatabaseType<Double>

	/**
	 * Descriptor of a raw binary data type for static type-checking.
	 * @see DatabaseTypesProvider.BINARY_DATA
	 */
	@OptIn(ExperimentalUnsignedTypes::class)
	public interface BINARY_DATA : DatabaseType<UByteArray> {
		/**
		 * Maximum count of bytes that can be stored by values of this type ot `null` if there is no limit.
		 */
		public val maxSizeOrNull: UInt?

		/**
		 * `true` if value can't hold less than [maxSizeOrNull][DatabaseType.BINARY_DATA.maxSizeOrNull] bytes.
		 */
		public val isFixedSize: Boolean
	}
}
