package ru.landgrafhomyak.serdha.api.v0.ddl.types

import kotlin.jvm.JvmName
import ru.landgrafhomyak.serdha.api.v0.ddl.table.Table

/**
 * Provider of database types descriptors.
 */
@Suppress("FunctionName", "PropertyName")
public interface DatabaseTypesProvider {
	/**
	 * Type of internal row id. Can't be casted to types like integers or pointers.
	 * @param TableUserExtension User's extension type for target table. For static type checking.
	 * @param table Descriptor of table whose row_id type will be obtained.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@JvmName("ROW_ID")
	public fun <TableUserExtension : Any> ROW_ID(table: Table<TableUserExtension, *>): DatabaseType<RowId<TableUserExtension>>

	/**
	 * Descriptor of a boolean type. Fields of this type can contain only 2 values: `true` and `false`.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("BOOLEAN")
	public val BOOLEAN: DatabaseType.BOOLEAN

	/**
	 * Descriptor of a signed integer type with size at least 8 bits. Can contain values in range `-128..127`.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("INT_S8")
	public val INT_S8: DatabaseType.INT_S8

	/**
	 * Descriptor of an unsigned integer type with size at least 8 bits. Can contain values in range `0..255`.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("INT_U8")
	public val INT_U8: DatabaseType.INT_U8

	/**
	 * Descriptor of a signed integer type with size at least 16 bits. Can contain values in range `-32768..32767`.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("INT_S16")
	public val INT_S16: DatabaseType.INT_S16

	/**
	 * Descriptor of an unsigned integer type with size at least 16 bits. Can contain values in range `0..65535`.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("INT_U16")
	public val INT_U16: DatabaseType.INT_U16

	/**
	 * Descriptor of a signed integer type with size at least 32 bits. Can contain values in range `-2147483648..2147483647`.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("INT_S32")
	public val INT_S32: DatabaseType.INT_S32

	/**
	 * Descriptor of an unsigned integer type with size at least 32 bits. Can contain values in range `0..4294967295`.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("INT_U32")
	public val INT_U32: DatabaseType.INT_U32

	/**
	 * Descriptor of a signed integer type with size at least 64 bits. Can contain values in range `-9223372036854775808..9223372036854775807`.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("INT_S64")
	public val INT_S64: DatabaseType.INT_S64

	/**
	 * Descriptor of an unsigned integer type with size at least 64 bits. Can contain values in range `0..18446744073709551615`.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("INT_U64")
	public val INT_U64: DatabaseType.INT_U64

	/**
	 * Descriptor of a local date type in some format like `(year, month, day)`. Day component can contain values in `1..31`, month in `1..12` and year in `-32768..32767`.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("LOCAL_DATE")
	public val LOCAL_DATE: DatabaseType.LOCAL_DATE

	/**
	 * Descriptor of a local time type in some format like `(hour, minute, second)`. Hour component can contain values in `0..23`, minute in `0..59` and second in `0..59`.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("LOCAL_TIME")
	public val LOCAL_TIME: DatabaseType.LOCAL_TIME

	/**
	 * Descriptor of union of types [LOCAL_DATE][DatabaseType.Builder.LOCAL_DATE] and [LOCAL_TIME][DatabaseType.Builder.LOCAL_TIME].
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("LOCAL_DATETIME")
	public val LOCAL_DATETIME: DatabaseType.LOCAL_DATETIME

	/**
	 * Descriptor of a timestamp type. Can contain values in `-9223372036854775808..9223372036854775807`
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("INSTANT")
	public val INSTANT: DatabaseType.INSTANT

	/**
	 * Descriptor of a single-precision floating-point number with size at least 32 bits. See [IEEE 754](https://en.wikipedia.org/wiki/IEEE_754).
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("FLOAT_32")
	public val FLOAT_32: DatabaseType.FLOAT_32

	/**
	 * Descriptor of a double-precision floating-point number with size at least 64 bits. See [IEEE 754](https://en.wikipedia.org/wiki/IEEE_754).
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("FLOAT_64")
	public val FLOAT_64: DatabaseType.FLOAT_64

	/**
	 * Descriptor of a raw binary data type without size restrictions (must be controlled by user).
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@get:JvmName("BINARY_DATA")
	public val BINARY_DATA: DatabaseType.BINARY_DATA

	/**
	 * Descriptor of a raw binary data type that can holds maximum [size] bytes.
	 *
	 * @param size Maximum count of bytes that can be stored in the value of this type.
	 * @param isFixedSize If `true`, the type will restrict values with less than [size] bytes.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@JvmName("BINARY_DATA")
	public fun BINARY_DATA(size: UInt, isFixedSize: Boolean): DatabaseType.BINARY_DATA


	/**
	 * Returns descriptor of a new type, which can contain all values of [original type][notNullType] or `null` value.
	 *
	 * Descriptors of nullable types are returned itself without any modifications.
	 */
	@Suppress("INAPPLICABLE_JVM_NAME")
	@JvmName("nullableOf")
	public fun <R : Any, D : DatabaseType<R>> nullableOf(notNullType: D): DatabaseType.Nullable<R, D>
}